use super::{Cookie, ResponseBody, Status};

type HeaderMap = Vec<String>;

#[derive(Debug)]
/// Enum for the result of a Handling Function, where...
///
/// [Outcome::Success] represents that the route
/// was successful and the Answer is contained in \[S\].
/// [Outcome::Failure] represents that it was unsuccessful and nobody else is going to be
/// successful. \[E\] represnts the Error Code.
/// [Outcome::Forward] represents that some requirements weren't met for a route to be working with
/// the request so the next one that matches should cover that \[F\] represents the maybe processed
/// data of the request.
///
/// # Example
/// ```
/// use http::handling::{response::{Outcome, Response, Status}, routes::Data, request::Request};
/// fn handler(request: Request, _data: Data) -> Outcome<Response, Status, Data> {
///     todo!()
/// }
/// ```
pub enum Outcome<S, E, F> {
    Success(S),
    Failure(E),
    Forward(F),
}

/// Response is a wrapper for http responses.
#[derive(Debug)]
pub struct Response {
    /// the [`Vec<String>`] of headers unrelated to `Content-Type` and `Content-Length`
    pub headers: HeaderMap,
    /// Optional Cookie in the response
    pub cookies: Option<Cookie>,
    /// Status code of the response
    pub status: Option<Status>,
    /// Response body and `Content-Type` and `Content-Length` headers.
    pub body: Box<dyn ResponseBody>,
}
