use std::str::FromStr;

use crate::{
    setup::MountPoint,
    utils::{
        url_utils::datatypes::RawUriElement,
        urlencoded::{EnCodable, UrlEncodeData},
        vec_utils::remove_n,
    },
};

use super::datatypes::{ParseUriError, RawUri, Uri};

impl MountPoint {
    pub fn compare_with_uri(&self, uri: &mut Uri) -> bool {
        if !uri.raw.starts_with(&self.mountpoint.raw) {
            return false;
        }
        let to_remove_after_finish = self.mountpoint.parts.len();
        remove_n(uri.mut_parts(), to_remove_after_finish);
        true
    }
}

impl Uri {
    pub fn new(parts: Vec<&str>) -> Self {
        Self {
            raw: "/".to_owned()
                + &parts
                    .iter()
                    .map(|part| part.encode())
                    .collect::<Vec<String>>()
                    .join("/"),
            parts: parts.into_iter().map(UrlEncodeData::from_raw).collect(),
        }
    }
    pub fn parts(&self) -> &[UrlEncodeData] {
        self.parts.as_slice()
    }
    pub fn mut_parts(&mut self) -> &mut Vec<UrlEncodeData> {
        self.parts.as_mut()
    }
    pub fn compare(&self, uri: &Uri) -> bool {
        let mut uri_iter = uri.parts.iter();
        for part in self.parts.iter() {
            let Some(part_uri) = uri_iter.next() else {
                return false;
            };
            if part != part_uri {
                return false;
            }
        }
        true
    }
    pub(crate) fn to_pretty_print_string(&self) -> String {
        if self.parts.is_empty() {
            "/".to_string()
        } else {
            let url = self
                .parts
                .iter()
                .map(|part| part.encoded())
                .collect::<Vec<_>>();
            "/".to_string() + &url.join("/")
        }
    }
}

impl RawUri {
    pub fn new(parts: Vec<&str>) -> Self {
        let mut result = Self {
            infinte_end: false,
            parts: Vec::with_capacity(parts.len()),
            raw_string: parts.join("/"),
        };
        for part in parts {
            if part.starts_with('<') && part.ends_with("..>") {
                result.infinte_end = true;
                break;
            }
            if part.starts_with('<') && part.ends_with('>') {
                result.parts.push(RawUriElement::Variable);
                continue;
            }
            result
                .parts
                .push(RawUriElement::Name(UrlEncodeData::from_raw(part)))
        }
        result
    }
    pub fn compare_uri(&self, uri: &Uri) -> bool {
        let mut iter_comp = uri.parts.iter();
        if uri.parts().len() != self.parts.len() && !self.infinte_end {
            return false;
        }
        for element in self.parts.iter() {
            let Some(compare_element) = iter_comp.next() else {
                return false;
            };
            if *element == RawUriElement::Variable {
                continue;
            }
            if let RawUriElement::Name(name) = element {
                if name.encoded() != compare_element.encoded() {
                    return false;
                }
            } else {
                return false;
            }
        }
        if uri.parts.len() > self.parts.len() && self.infinte_end {
            return true;
        }
        true
    }
    pub(crate) fn to_pretty_print_string(&self, is_after: &Uri) -> String {
        let is_after = is_after.to_pretty_print_string();
        if is_after == "/" {
            format!("\x1b[34;4m/\x1b[24m{}\x1b[0m", self)
        } else {
            format!("\x1b[34;4m{is_after}/\x1b[24m{self}\x1b[0m")
        }
    }
}

impl std::fmt::Display for RawUri {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.raw_string)
    }
}

impl std::fmt::Display for Uri {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if self.parts.is_empty() {
            write!(f, "/")
        } else {
            let url = self
                .parts
                .iter()
                .map(|part| part.encoded())
                .collect::<Vec<_>>();
            write!(f, "{}", url.join("/"))
        }
    }
}

impl FromStr for Uri {
    type Err = ParseUriError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let split = s.split('/');
        let mut result = Vec::new();
        for sub in split {
            if sub.is_empty() {
                continue;
            }
            result.push(if let Ok(coded) = UrlEncodeData::from_encoded(sub) {
                coded
            } else {
                UrlEncodeData::from_raw(sub)
            });
        }
        Ok(Self {
            parts: result,
            raw: s.to_string(),
        })
    }
}

impl FromStr for RawUri {
    type Err = ParseUriError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts = s.split('/').collect::<Vec<&str>>();
        let mut result = Self {
            infinte_end: false,
            parts: Vec::new(),
            raw_string: s.to_string(),
        };
        for part in parts {
            if part.is_empty() {
                continue;
            }
            if part.starts_with('<') && part.ends_with("..>") {
                result.infinte_end = true;
                break;
            }
            if part.starts_with('<') && part.ends_with('>') {
                result.parts.push(RawUriElement::Variable);
                continue;
            }
            result
                .parts
                .push(RawUriElement::Name(UrlEncodeData::from_raw(part)))
        }
        println!("{:?}", result);
        Ok(result)
    }
}

impl TryFrom<&str> for Uri {
    type Error = ParseUriError;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Self::from_str(value)
    }
}

impl TryFrom<&str> for RawUri {
    type Error = ParseUriError;
    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Self::from_str(value)
    }
}

#[cfg(test)]
mod test {
    use crate::utils::{
        url_utils::{
            datatypes::{RawUri, RawUriElement},
            Uri,
        },
        urlencoded::UrlEncodeData,
    };

    #[test]
    fn uri_to_string() {
        assert_eq!("a/%20", Uri::new(vec!["a", " "]).to_string());
    }

    #[test]
    fn uri_from_string() {
        assert_eq!(Uri::new(vec!["a", " "]), "/a/%20".parse().unwrap())
    }

    #[test]
    fn raw_uri_from_string() {
        assert_eq!(
            RawUri {
                raw_string: "/<name>/a/<f..>".to_owned(),
                infinte_end: true,
                parts: vec![
                    RawUriElement::Variable,
                    RawUriElement::Name(UrlEncodeData::from_raw("a")),
                ]
            },
            "/<name>/a/<f..>".parse().unwrap()
        );
    }

    #[test]
    fn raw_uri_to_string() {
        assert_eq!(
            "/<name>/a/<f..>",
            RawUri {
                raw_string: "/<name>/a/<f..>".to_owned(),
                infinte_end: true,
                parts: vec![
                    RawUriElement::Variable,
                    RawUriElement::Name(UrlEncodeData::from_raw("a")),
                ]
            }
            .to_string()
        )
    }
}
